# Copyright (C) Cardiff University (2018-2021)
# SPDX-License-Identifier: MIT

"""
`gwosc.locate` provides functions to determine the file URLs containing
data for a specific dataset.

You can search for remote data URLS based on the event name:

>>> from gwosc.locate import get_event_urls
>>> get_event_urls("GW150914")
['https://gwosc.org/eventapi/json/GWTC-1-confident/GW150914/v3/H-H1_GWOSC_4KHZ_R1-1126259447-32.hdf5',
 'https://gwosc.org/eventapi/json/GWTC-1-confident/GW150914/v3/H-H1_GWOSC_4KHZ_R1-1126257415-4096.hdf5',
 'https://gwosc.org/eventapi/json/GWTC-1-confident/GW150914/v3/L-L1_GWOSC_4KHZ_R1-1126259447-32.hdf5',
 'https://gwosc.org/eventapi/json/GWTC-1-confident/GW150914/v3/L-L1_GWOSC_4KHZ_R1-1126257415-4096.hdf5']

You can down-select the URLs using keyword arguments:

>>> get_event_urls("GW150914", detector="L1", duration=32)
['https://gwosc.org/eventapi/json/GWTC-1-confident/GW150914/v3/L-L1_GWOSC_4KHZ_R1-1126259447-32.hdf5']

You can search for remote data URLs based on the GPS time interval as
follows:

>>> from gwosc.locate import get_urls
>>> get_urls("L1", 968650000, 968660000)
['https://gwosc.org/archive/data/S6/967835648/L-L1_LOSC_4_V1-968646656-4096.hdf5',
 'https://gwosc.org/archive/data/S6/967835648/L-L1_LOSC_4_V1-968650752-4096.hdf5',
 'https://gwosc.org/archive/data/S6/967835648/L-L1_LOSC_4_V1-968654848-4096.hdf5',
 'https://gwosc.org/archive/data/S6/967835648/L-L1_LOSC_4_V1-968658944-4096.hdf5']

By default, this method will return the paths to HDF5 files for the 4 kHz
sample-rate data, these can be specified as keyword arguments.
For full information, see :func:`get_urls`.
"""  # noqa: E501

from . import datasets, utils
from .api import DEFAULT_URL
from .api import v2 as apiv2

__all__ = ["get_urls", "get_run_urls", "get_event_urls"]
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def get_urls(
    detector,
    start,
    end,
    dataset=None,
    version=None,
    sample_rate=4096,
    format="hdf5",
    host=DEFAULT_URL,
):
    """Fetch the URLs from GWOSC regarding a given GPS interval

    Parameters
    ----------
    detector : `str`
        the prefix of the relevant GW detector

    start : `int`
        the GPS start time of your query

    end : `int`
        the GPS end time of your query

    dataset : `str`, `None`, optional
        the name of the dataset to query, e.g. ``'GW150914'``

    version : `int`, `None`, optional
        the data-release version for the selected datasets

    sample_rate : `int`, optional, default : ``4096``
        the sampling rate (Hz) of files you want to find

    format : `str`, optional, default: ``'hdf5'``
        the file format (extension) you want to find

    host : `str`, optional
        the URL of the remote GWOSC server

    Returns
    -------
    urls : `list` of `str`
        the list of remote file URLs that contain data matching the
        relevant parameters
    """
    start = int(start)
    end = int(end)

    for dstype in ("event", "run"):
        dsets = datasets._iter_datasets(
            match=dataset,
            type=dstype,
            detector=detector,
            segment=(start, end),
            version=version,
            host=host,
        )

        for dst in dsets:
            # get URL list for this dataset
            if dstype == "run":
                urls = get_run_urls(
                    dst,
                    detector,
                    start,
                    end,
                    host=host,
                    sample_rate=sample_rate,
                    format=format,
                )
            else:
                urls = get_event_urls(
                    dst,
                    detector=detector,
                    start=start,
                    end=end,
                    format=format,
                    sample_rate=sample_rate,
                    version=version,
                    host=host,
                )

            # if full span covered, return now
            if utils.full_coverage(urls, (start, end)):
                return urls

    raise ValueError(
        f"Cannot find a GWOSC dataset for {detector} covering [{start}, {end})"
    )


def get_run_urls(
    run,
    detector,
    start,
    end,
    format="hdf5",
    sample_rate=4096,
    host=DEFAULT_URL,
    **match,
):
    """Fetch the URLs from GWOSC regarding a given event

    Parameters
    ----------
    run : `str`
        the ID of the run

    detector : `str`, optional
        the detector for files you want to find

    start : `int`
        the GPS start time of your query

    end : `int`
        the GPS end time of your query

    format : `str`, optional, default: ``'hdf5'``
        the file format (extension) you want to find

    sample_rate : `int`, optional, default : ``4096``
        the sampling rate (Hz) of files you want to find

    host : `str`, optional
        the URL of the remote GWOSC server

    Returns
    -------
    urls : `list` of `str`
        the list of remote file URLs that contain data matching the
        relevant parameters
    """
    return [
        afile[f"{format}_url"]
        for afile in apiv2.fetch_run_strain_files(
            run=run,
            start=start,
            end=end,
            detector=detector,
            sample_rate=sample_rate,
            host=host,
        )
    ]


def get_event_urls(
    event,
    catalog=None,
    version=None,
    detector=None,
    start=None,
    end=None,
    format="hdf5",
    sample_rate=4096,
    host=DEFAULT_URL,
    **match,
):
    """Fetch the URLs from GWOSC regarding a given event

    Parameters
    ----------
    event : `str`
        the ID of the event

    detector : `str`, optional
        the detector for files you want to find

    format : `str`, optional, default: ``'hdf5'``
        the file format (extension) you want to find

    sample_rate : `int`, optional, default : ``4096``
        the sampling rate (Hz) of files you want to find

    host : `str`, optional
        the URL of the remote GWOSC server

    start : `int`
        the GPS start time of your query

    end : `int`
        the GPS end time of your query

    version : `int`, `None`, optional
        the data-release version for the selected datasets

    catalog : `str`, `None`, optional
        the name of the catalog for the selected datasets

    Returns
    -------
    urls : `list` of `str`
        the list of remote file URLs that contain data matching the
        relevant parameters
    """
    # get the URL file list for this event
    strain_files = list(
        apiv2.fetch_event_strain_data(
            event,
            version=version,
            catalog=catalog,
            detector=detector,
            sample_rate=sample_rate,
            duration=match.get("duration"),
            format=format,
            host=host,
        )
    )

    # If there are no strain files for this event, return run strain files
    if len(strain_files) == 0:
        event_obj = apiv2.fetch_event_version(
            event, catalog=catalog, version=version, host=host
        )
        gps = event_obj["gps"]
        return get_run_urls(
            event_obj.get("run"),
            detector,
            gps,
            gps,
            format=format,
            sample_rate=sample_rate,
            host=host,
        )

    # format start and end as a segment
    if start is None and end is None:
        segment = None
    else:
        segment = (
            -float("inf") if start is None else start,
            +float("inf") if end is None else end,
        )

    strain_urls = []
    # filter on the requested segment
    for urlmeta in strain_files:
        if segment:  # check overlap
            _start = urlmeta["gps_start"]
            thisseg = (_start, _start + urlmeta["duration"])
            if not utils.segments_overlap(segment, thisseg):
                continue
        strain_urls.append(urlmeta["download_url"])

    return strain_urls



# Copyright (C) Cardiff University (2018-2021)
# SPDX-License-Identifier: MIT

"""Tests for :mod:`gwosc.urls`"""

import pytest

from .. import urls as gwosc_urls

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


@pytest.mark.remote
def test_sieve(gw150914_strain):
    nfiles = len(gw150914_strain)
    sieved = list(
        gwosc_urls.sieve(
            gw150914_strain,
            detector="L1",
        )
    )
    assert len(sieved) == nfiles // 2
    sieved = list(
        gwosc_urls.sieve(
            gw150914_strain,
            detector="L1",
            sampling_rate=4096,
        )
    )
    assert len(sieved) == nfiles // 4


# -- local tests


def test_sieve_local(mock_strain):
    assert (
        list(
            gwosc_urls.sieve(mock_strain, detector="X1"),
        )
        == mock_strain[:2]
    )
    assert (
        list(
            gwosc_urls.sieve(mock_strain, sampling_rate=16384),
        )
        == mock_strain[3:]
    )


def test_sieve_error(mock_strain):
    with pytest.raises(TypeError):
        list(gwosc_urls.sieve(mock_strain, blah=1))


def test_match_local(mock_urls):
    assert (
        gwosc_urls.match(
            mock_urls,
            tag="TEST",
            start=0,
            end=1,
        )
        == mock_urls[:1]
    )
    assert (
        gwosc_urls.match(
            mock_urls,
            sample_rate=16384,
        )
        == mock_urls[3:]
    )
    assert (
        gwosc_urls.match(
            mock_urls,
            tag="TEST2",
            version="R2",
            sample_rate=16384,
        )
        == mock_urls[3:]
    )


def test_match_local_tag_error(mock_urls):
    with pytest.raises(ValueError) as exc:
        gwosc_urls.match(mock_urls)
    assert str(exc.value).startswith("multiple GWOSC URL tags")


def test_match_empty():
    assert gwosc_urls.match([]) == []



from gwosc import __version__ as gwosc_version

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
needs_sphinx = "1.6"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx_automodapi.automodapi",
    "sphinx_immaterial_igwn",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "GWOSC"
copyright = "2018-2020, Cardiff University"
author = "Duncan Macleod"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = gwosc_version.split(".dev", 1)[0]
# The full version, including alpha/beta/rc tags.
release = gwosc_version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The reST default role (used for this markup: `text`) to use for all
# documents.
default_role = "obj"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_immaterial_igwn"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "edit_uri": "-/tree/main/docs",
    "repo_name": "gwosc client",
    "repo_url": "https://git.ligo.org/gwosc/client",
    "icon": {
        "repo": "fontawesome/brands/gitlab",
    },
    "site_url": "https://gwosc.readthedocs.io",
}
html_last_updated_fmt = "%Y-%m-%d"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]

# -- Extensions -----------------------------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "requests": ("https://requests.readthedocs.io/en/stable/", None),
}

# Write automodsumm files to /docs/reference/
automodapi_toctreedirnm = "reference"
automodapi_inherited_members = False
automodsumm_inherited_members = False



# Copyright (C) Cardiff University (2018-2021)
# SPDX-License-Identifier: MIT

"""Tests for :mod:`gwosc.timeline`"""

from unittest import mock

import pytest

from .. import timeline

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


@pytest.mark.remote
@pytest.mark.parametrize(
    "flag, start, end, result",
    [
        (
            "H1_DATA",
            1126051217,
            1126151217,
            [
                (1126073529, 1126114861),
                (1126121462, 1126123267),
                (1126123553, 1126126832),
                (1126139205, 1126139266),
                (1126149058, 1126151217),
            ],
        ),
        (
            "L1_DATA",
            1126259446,
            1126259478,
            [
                (1126259446, 1126259478),
            ],
        ),
    ],
)
def test_get_segments(flag, start, end, result):
    assert timeline.get_segments(flag, start, end) == result


@pytest.mark.remote
def test_get_segments_long():
    assert len(timeline.get_segments("H1_DATA", 1126051217, 1137196817)) == 654


@pytest.mark.remote
def test_timeline_url():
    # check that unknown IFO results in no matches
    with pytest.deprecated_call():
        with pytest.raises(ValueError):
            timeline.timeline_url("X1", 1126259446, 1126259478)


@mock.patch("gwosc.timeline._find_dataset", return_value="S6")
def test_timeline_url_local(find):
    with pytest.deprecated_call():
        assert timeline.timeline_url("L1_DATA", 0, 1, host="test") == (
            "test/timeline/segments/json/S6/L1_DATA/0/1/"
        )
    find.assert_called_with(0, 1, "L1", host="test")



# Copyright (C) Cardiff University (2018-2021)
# SPDX-License-Identifier: MIT

"""Utilities for URL handling"""

import re
from os.path import basename, splitext

from .utils import segments_overlap

# GWOSC filename re
URL_REGEX = re.compile(
    r"\A((.*/)*(?P<obs>[^/]+)-"
    r"(?P<ifo>[A-Z][0-9])_(L|GW)OSC_"
    r"((?P<tag>[^/]+)_)?"
    r"(?P<samp>\d+(KHZ)?)_"
    r"[RV](?P<version>\d+)-"
    r"(?P<strt>[^/]+)-"
    r"(?P<dur>[^/\.]+)\."
    r"(?P<ext>[^/]+))\Z"
)

VERSION_REGEX = re.compile(r"[RV]\d+")


def sieve(urllist, segment=None, **match):
    """Sieve a list of GWOSC URL metadata dicts based on key, value pairs

    Parameters
    ----------
    urllist : `list` of `dict`
        the ``'strain'`` metadata list, as retrieved from the GWOSC
        server

    segment : `tuple` of `int`
        a ``[start, stop)`` GPS segment against which to check overlap
        for each URL

    **match
        other keywords match **exactly** against the corresponding key
        in the `dict` for each URL

    Yields
    ------
    dict :
        each URL dict that matches the parameters is yielded, in the same
        order as the input ``urllist``
    """
    # remove null keys
    match = {key: value for key, value in match.items() if value is not None}

    # sieve
    for urlmeta in urllist:
        try:
            if any(match[key] != urlmeta[key] for key in match):
                continue
        except KeyError as exc:
            raise TypeError(f"unrecognised match parameter: {str(exc)}")
        if segment:  # check overlap
            _start = urlmeta["GPSstart"]
            thisseg = (_start, _start + urlmeta["duration"])
            if not segments_overlap(segment, thisseg):
                continue
        yield urlmeta


def _match_url(
    url,
    detector=None,
    start=None,
    end=None,
    tag=None,
    sample_rate=None,
    version=None,
    duration=None,
    ext=None,
):
    """Match a URL against requested parameters

    Returns
    -------
    None
        if the URL doesn't match the request

    tag, version : `str`, `int`
        if the URL matches the request

    Raises
    ------
    StopIteration
        if the start time of the URL is _after_ the end time of the
        request
    """
    reg = URL_REGEX.match(basename(url)).groupdict()
    for param, regvar in (
        (detector, reg["ifo"]),
        (tag, reg["tag"]),
        (version, int(reg["version"])),
        (sample_rate, float(reg["samp"].rstrip("KHZ")) * 1024),
        (duration, float(reg["dur"]) != duration),
        (ext, reg["ext"] != ext),
    ):
        # if param was given and the regex value doesn't match, ignore this url
        if param and regvar != param:
            return

    # match times
    if end is not None:
        gps = int(reg["strt"])
        if gps >= end:  # too late
            return

    if start is not None:
        gps = int(reg["strt"])
        dur = int(reg["dur"])
        if gps + dur <= start:  # too early
            return

    return reg["tag"], int(reg["version"])


def match(
    urls,
    detector=None,
    start=None,
    end=None,
    tag=None,
    sample_rate=None,
    version=None,
    duration=None,
    ext=None,
):
    """Match GWOSC URLs for a given [start, end) interval

    Parameters
    ----------
    urls : `list` of `str`
        List of URL paths

    start : `int`
        GPS start time of match interval

    end : `int`
        GPS end time of match interval

    tag : `str`, optional
        URL tag to match, e.g. ``'CLN'``

    version : `int`, optional
        Data release version to match, defaults to highest available
        version

    Returns
    -------
    urls : `list` of `str`
        A sub-list of the input, based on matching, if no URLs are matched,
        the return will be empty ``[]``.
    """
    matched = {}
    matched_tags = set()

    # sort URLs by duration, then start time, then ...
    urls = sorted(
        urls,
        key=lambda u: splitext(basename(u))[0].split("-")[::-1],
    )

    # format version request
    if VERSION_REGEX.match(str(version)):
        version = version[1:]
    if version is not None:
        version = int(version)

    # loop URLS
    for url in urls:
        m = _match_url(
            url,
            detector=detector,
            start=start,
            end=end,
            tag=tag,
            sample_rate=sample_rate,
            version=version,
            duration=duration,
            ext=ext,
        )
        if m is None:
            continue

        mtag, mvers = m
        matched_tags.add(mtag)
        matched.setdefault(mvers, [])
        matched[mvers].append(url)

    # if multiple file tags found, and user didn't specify, error
    if len(matched_tags) > 1:
        tags = ", ".join(map(repr, matched_tags))
        raise ValueError(
            "multiple GWOSC URL tags discovered in dataset, "
            f"please select one of: {tags}"
        )

    # extract highest version
    try:
        return matched[max(matched)]
    except ValueError:  # no matched files
        return []



# Copyright (C) Cardiff University (2018-2021)
# SPDX-License-Identifier: MIT

"""
`gwosc.datasets` includes functions to query for available datasets.

To search for all available datasets:

>>> from gwosc import datasets
>>> datasets.find_datasets()
['GW150914', 'GW151226', 'GW170104', 'GW170608', 'GW170814', 'GW170817', 'LVT151012', 'O1', 'S5', 'S6']
>>> datasets.find_datasets(detector="V1")
['GW170814', 'GW170817']
>>> datasets.find_datasets(type="run")
['O1', 'S5', 'S6']

To query for the GPS time of an event dataset (or vice-versa):

>>> datasets.event_gps("GW170817")
1187008882.43
>>> datasets.event_at_gps(1187008882)
'GW170817'

Similar queries are available for observing run datasets:

>>> datasets.run_segment("O1")
(1126051217, 1137254417)
>>> datasets.run_at_gps(1135136350)  # event_gps('GW151226')
'O1'

To run an event query filtered by merger parameters:

>>> from gwosc.datasets import query_events
>>> query_events(select=["mass-1-source <= 3.0"])
['GW170817-v3', 'GW190425-v1', 'GW190425-v2', 'GW190425_081805-v3']
"""  # noqa: E501

import re
import warnings

from . import (
    api,
    urls,
    utils,
)
from .api import DEFAULT_URL
from .api import v2 as apiv2

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

_IGNORE = {
    "tenyear",
    "history",
    "oldhistory",
}


def _match_dataset(targetdetector, detectors, targetsegment, segment):
    """Returns `True` if the dataset metadata matches the target"""
    warnings.warn(
        "_match_dataset function will be removed in future versions.",
        DeprecationWarning,
        stacklevel=2,
    )
    if targetdetector not in set(detectors) | {None}:
        return False
    if targetsegment is None or utils.segments_overlap(targetsegment, segment):
        return True


def _run_datasets(detector=None, segment=None, host=DEFAULT_URL):
    runs = apiv2.fetch_runs(host=host)
    for run in runs:
        # Skip if detector was given and it doesn't match
        if detector is not None and detector not in run["detectors"]:
            continue
        # Skip if segment was given and it doesn't overlap
        if segment is not None:
            run_segment = (run["gps_start"], run["gps_end"])
            do_overlap = utils.segments_overlap(run_segment, segment)
            if not do_overlap:
                continue
        yield run["name"]


def _catalog_datasets(host=DEFAULT_URL):
    for catalog in apiv2.fetch_catalogs(host=host):
        yield catalog["name"]


def _match_event_dataset(
    dataset,
    catalog=None,
    version=None,
    detector=None,
    segment=None,
    host=DEFAULT_URL,
):
    warnings.warn(
        "_match_event_dataset function will be removed in future versions.",
        DeprecationWarning,
        stacklevel=2,
    )
    # get strain file list (matching catalog and version)
    full = bool(detector or segment)
    try:
        meta = _event_metadata(
            dataset,
            catalog=catalog,
            version=version,
            full=full,
            host=host,
        )
    except ValueError:  # no dataset matching catalog and/or version
        return False

    if not full:  # detector=None, segment=None
        return True
    try:
        strain = meta["strain"]
    except KeyError:  # pragma: no cover
        # no strain file list for this dataset
        return False

    # match detector
    if detector not in {None} | {u["detector"] for u in strain}:
        return False

    # match segment
    if segment is None:
        return True
    if not strain:
        return False
    eseg = utils.strain_extent(urls.sieve(strain, detector=detector))
    return utils.segments_overlap(segment, eseg)


def _event_datasets(
    detector=None,
    segment=None,
    catalog=None,
    version=None,
    host=DEFAULT_URL,
):
    # This segment expansion is needed to account for segments
    # that do not contain events, but are contained in the
    # 4096s event dataset (strain file)
    if segment is not None:
        start, end = segment
        expanded_segment = (start, end + 4096)
    else:
        expanded_segment = None

    events = apiv2.fetch_event_versions(
        segment=expanded_segment, catalogs=catalog, host=host
    )
    if version is not None:
        # Filter by version
        events = filter(lambda e: e["version"] == version, events)
    if detector is not None:
        # Filter by detector
        events = filter(lambda e: detector in e["detectors"], events)

    def _rank_catalog(cat):
        if "confident" in cat:
            return 1
        for word in ("marginal", "preliminary"):
            if word in cat:
                return 10
        return 5

    def _gps_distance(gps):
        if segment is None:
            return 0
        return int(abs(segment[0] - gps))

    for event in sorted(
        events,
        key=lambda e: (
            _gps_distance(e["gps"]),
            _rank_catalog(e["catalog"].lower()),
            -e["version"],
        ),
    ):
        yield f"{event['shortName']}"


def _iter_datasets(
    detector=None,
    type=None,
    segment=None,
    catalog=None,
    version=None,
    match=None,
    host=DEFAULT_URL,
):
    # get queries
    type = str(type).rstrip("s").lower()
    needruns = type in {"none", "run"}
    needcatalogs = type in {"none", "catalog"}
    needevents = type in {"none", "event"}

    # warn about event-only keywords when not querying for events
    if not needevents:
        for key, value in dict(catalog=catalog, version=version).items():
            if value is not None:
                warnings.warn(
                    f"the '{key}' keyword is only relevant when querying "
                    "for event datasets, it will be ignored now",
                )

    if match:
        reg = re.compile(match)

    def _yield_matches(iter_):
        for x in iter_:
            if not match or reg.search(x):
                yield x

    # search for events and datasets
    for needthis, collection in (
        (
            needruns,
            _run_datasets(
                detector=detector,
                host=host,
                segment=segment,
            ),
        ),
        (
            needcatalogs,
            _catalog_datasets(
                host=host,
            ),
        ),
        (
            needevents,
            _event_datasets(
                detector=detector,
                segment=segment,
                host=host,
                version=version,
                catalog=catalog,
            ),
        ),
    ):
        if not needthis:
            continue
        yield from _yield_matches(collection)


def find_datasets(
    detector=None,
    type=None,
    segment=None,
    match=None,
    catalog=None,
    version=None,
    host=DEFAULT_URL,
):
    """Find datasets available on the given GW open science host

    Parameters
    ----------
    detector : `str`, optional
        prefix of GW detector

    type : `str`, optional
        type of datasets to restrict, one of ``'run'``, ``'event'``, or
        ``'catalog'``

    segment : 2-`tuple` of `int`, `None`, optional
        a GPS ``[start, stop)`` interval to restrict matches to;
        datasets will match if they overlap at any point; this is
        not used to filter catalogs

    match : `str`, optional
        regular expression string against which to match datasets

    host : `str`, optional
        the URL of the GWOSC host to query, defaults to
        https://gwosc.org

    Returns
    -------
    datasets : `list` of `str`
        the names of all matched datasets, possibly empty

    Examples
    --------
    (Correct as of 2018-03-14)

    >>> from gwosc.datasets import find_datasets
    >>> find_datasets()
    ['GW150914', 'GW151226', 'GW170104', 'GW170608', 'GW170814', 'GW170817',
     'LVT151012', 'O1', 'S5', 'S6']
    >>> find_datasets(detector="V1")
    ['GW170814', 'GW170817']
    >>> find_datasets(type="event")
    ['GW150914', 'GW151226', 'GW170104', 'GW170608', 'GW170814', 'GW170817',
     'LVT151012']
    """
    return sorted(
        list(
            _iter_datasets(
                detector=detector,
                type=type,
                segment=segment,
                catalog=catalog,
                version=version,
                match=match,
                host=host,
            )
        )
    )


# -- event utilities ----------------------------------------------------------


def _event_metadata(
    event,
    catalog=None,
    version=None,
    full=True,
    host=DEFAULT_URL,
):
    warnings.warn(
        "_event_metadata function will be removed in future versions.",
        DeprecationWarning,
        stacklevel=2,
    )
    return list(
        api._fetch_allevents_event_json(
            event,
            catalog=catalog,
            version=version,
            full=full,
            host=host,
        )["events"].values()
    )[0]


def event_gps(event, catalog=None, version=None, host=DEFAULT_URL):
    """Returns the GPS time of an open-data event

    Parameters
    ----------
    event : `str`
        the name of the event to query

    catalog : `str`, optional
        name of catalogue that hosts this event

    version : `int`, `None`, optional
        the version of the data release to use,
        defaults to the highest available version

    host : `str`, optional
        the URL of the GWOSC host to query, defaults to
        https://gwosc.org

    Returns
    -------
    gps : `float`
        the GPS time of this event

    Examples
    --------
    >>> from gwosc.datasets import event_gps
    >>> event_gps("GW170817")
    1187008882.43
    >>> event_gps("GW123456")
    ValueError: no event dataset found for 'GW123456'
    """
    event = apiv2.fetch_event_version(
        event, catalog=catalog, version=version, host=host
    )
    return event["gps"]


def event_segment(
    event,
    detector=None,
    catalog=None,
    version=None,
    host=DEFAULT_URL,
):
    """Returns the GPS ``[start, stop)`` interval covered by an event dataset

    Parameters
    ----------
    event : `str`
        the name of the event

    detector : `str`, optional
        prefix of GW detector

    catalog : `str`, optional
        name of catalogue that hosts this event

    version : `int`, `None`, optional
        the version of the data release to use,
        defaults to the highest available version

    host : `str`, optional
        the URL of the GWOSC host to query, defaults to
        https://gwosc.org

    Returns
    -------
    start, end : `int`
        the GPS ``[start, end)`` interval covered by this event dataset

    Examples
    --------
    >>> from gwosc.datasets import event_segment
    >>> event_segment("GW150914")
    segment(1126257415, 1126261511)
    """

    strain_files = apiv2.fetch_event_strain_data(
        event, detector=detector, version=None, catalog=None, host=host
    )

    if not strain_files:  # pragma: no cover
        raise ValueError(
            f"event '{event}' has no strain files",
        )

    # Return the strain files extent (mod of utils.strain_extent)
    starts, ends = zip(
        *[
            (afile["gps_start"], afile["gps_start"] + afile["duration"])
            for afile in strain_files
        ]
    )

    return min(starts), max(ends)


def event_at_gps(gps, host=DEFAULT_URL, tol=1):
    """Returns the name of the open-data event matching the GPS time

    This function will return the first event for which
    ``|eventgps - gps| < = tol``.

    Parameters
    ----------
    gps : `float`
        The GPS time to locate

    host : `str`, optional
        the URL of the GWOSC host to query, defaults to
        https://gwosc.org

    tol : `float`, optional
        the search window (in seconds), default: ``1``

    Returns
    -------
    event : `str`
        the name of the matched event

    Raises
    -------
    ValueError
        if no events are found matching the GPS time

    Examples
    --------
    >>> from gwosc.datasets import event_at_gps
    >>> event_at_gps(1187008882)
    'GW170817'
    >>> event_at_gps(1187008882, tol=0.1)
    ValueError: no event found within 0.1 seconds of 1187008882
    """
    events = list(apiv2.fetch_event_versions(segment=(gps - tol, gps + tol)))
    if len(events) < 1:
        raise ValueError(f"no event found within {tol} seconds of {gps}")
    return events[0]["name"]


def event_detectors(event, catalog=None, version=None, host=DEFAULT_URL):
    """Returns the `set` of detectors that observed an event

    Parameters
    ----------
    event : `str`
        the name of the event to query

    version : `int`, `None`, optional
        the version of the data release to use,
        defaults to the highest available version

    catalog : `str`, optional
        name of catalogue that hosts this event

    host : `str`, optional
        the URL of the GWOSC host to query, defaults to
        https://gwosc.org

    Returns
    -------
    detectors : `set`
        the set of detectors for which data file URLs are included in
        the data release

    Examples
    --------
    >>> from gwosc.datasets import event_detectors
    >>> event_detectors("GW150914")
    {'H1', 'L1'}
    """
    event = apiv2.fetch_event_version(
        event, catalog=catalog, version=version, host=host
    )
    return set(event["detectors"])


# -- run utilities ------------------------------------------------------------


def run_segment(run, host=DEFAULT_URL):
    """Returns the GPS ``[start, stop)`` interval covered by a run dataset

    Parameters
    ----------
    run : `str`
        the name of the run dataset to query

    host : `str`, optional
        the URL of the GWOSC host to query, defaults to
        https://gwosc.org

    Returns
    -------
    start, end : `int`
        the GPS ``[start, end)`` interval covered by this run dataset

    Examples
    --------
    >>> from gwosc.datasets import run_segment
    >>> run_segment("O1")
    segment(1126051217, 1137254417)
    >>> run_segment("Oh dear")
    ValueError: Run 'Oh dear' not found.
    """
    run = apiv2.fetch_run(run)
    return run["gps_start"], run["gps_end"]


def run_at_gps(gps, host=DEFAULT_URL):
    """Returns the name of the open-data run dataset matching the GPS time

    This function will return the first event for which
    ``start <= gps < end``

    Parameters
    ----------
    gps : `float`
        The GPS time to locate

    host : `str`, optional
        the URL of the GWOSC host to query, defaults to
        https://gwosc.org

    Returns
    -------
    run : `str`
        the name of the matched observing run

    Raises
    -------
    ValueError
        if no datasets are found matching the GPS time

    Examples
    --------
    >>> from gwosc.datasets import run_at_gps
    >>> run_at_gps(1135136350)
    'O1'
    >>> run_at_gps(0)
    ValueError: no run dataset found containing GPS 0
    """
    for run in apiv2.fetch_runs(host=host):
        start, end = run["gps_start"], run["gps_end"]
        if start <= gps < end:
            return run["name"]
    raise ValueError(f"no run dataset found containing GPS {gps}")


def dataset_type(dataset, host=DEFAULT_URL):
    """Returns the type of the named dataset

    Parameters
    ----------
    dataset : `str`
        The name of the dataset to match

    host : `str`, optional
        the URL of the GWOSC host to query

    Returns
    -------
    type : `str`
        the type of the dataset, one of 'run', 'event', or 'catalog'

    Raises
    -------
    ValueError
        if this dataset cannot be matched

    Examples
    --------
    >>> from gwosc.datasets import dataset_type
    >>> dataset_type("O1")
    'run'
    """
    for type_ in ("run", "catalog", "event"):
        if dataset in find_datasets(type=type_, host=host):
            return type_
    raise ValueError(
        f"failed to determine type for dataset {dataset!r}",
    )


def query_events(select, host=DEFAULT_URL):
    """Return a list of events filtered by the parameters in `select`

    Parameters
    ----------
    select : `list` of `str`
        A list of strings where each element is a range constrain on the
        event parameters.
        All ranges should have inclusive ends (<= and => operators).

    host : `str`, optional
        the URL of the GWOSC host to query

    Examples
    --------
    >>> from gwosc.datasets import query_events
    >>> query_events(
    ...     select=[
    ...         "mass-1-source >= 1.4",
    ...         "200 >= luminosity-distance >= 100",
    ...     ]
    ... )
    ['GW190425-v1', 'GW190425-v2', 'GW190425_081805-v3']

    Notes
    -----
    Operators:

    - `<=` (or `=<`)
    - `=>` (or `>=`)

    Parameters:

    - ``gps-time`` [s],
    - ``mass-1-source`` [solar mass],
    - ``mass-2-source`` [solar mass],
    - ``network-matched-filter-snr``,
    - ``luminosity-distance`` [Mpc],
    - ``chi-eff``,
    - ``total-mass-source`` [solar mass],
    - ``chirp-mass`` [solar mass],
    - ``chirp-mass-source`` [solar mass],
    - ``redshift``,
    - ``far`` [events/year],
    - ``p-astro``,
    - ``final-mass-source`` [solar mass],

    For a full description of all parameters see
    https://www.gwosc.org/apidocs/#event5
    """
    select_query = dict(utils.select_to_query(select, host=host))
    events = apiv2.fetch_event_versions(select=select_query, host=host)
    return [f"{e['shortName']}" for e in events]



# Copyright (C) Cardiff University (2018-2021)
# SPDX-License-Identifier: MIT

"""Tests for :mod:`gwosc.api`"""

from unittest import mock

import pytest
from requests import RequestException

from ..api import v1 as api

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


_ALLOWED_PARAMS = [
    "gps-time",
    "mass-1-source",
    "mass-2-source",
    "network-matched-filter-snr",
    "luminosity-distance",
    "chi-eff",
    "total-mass-source",
    "chirp-mass",
    "chirp-mass-source",
    "redshift",
    "far",
    "p-astro",
    "final-mass-source",
]


def losc_url(path):
    return f"{api.DEFAULT_URL}/{path}"


def check_json_url_list(urllist, keys={"detector", "format", "url"}):
    assert isinstance(urllist, list)
    for urld in urllist:
        for key in keys:
            assert key in urld


@pytest.mark.remote
def test_fetch_json():
    url = "https://gwosc.org/archive/1126257414/1126261510/json/"
    out = api.fetch_json(url)
    assert isinstance(out, dict)
    assert len(out["events"]) == 3
    assert sorted(out["events"]["GW150914-v1"]["detectors"]) == ["H1", "L1"]
    assert {"O1", "O1_16KHZ", "history"}.issubset(set(out["runs"]))


@pytest.mark.remote
def test_fetch_json_error():
    # check errors (use legit URL that isn't JSON)
    url = "https://gwosc.org/archive/1126257414/1126261510/"
    with pytest.raises((RequestException, ValueError)):
        api.fetch_json(url)


def test_fetch_json_local(requests_mock):
    url = "http://anything"
    requests_mock.get(
        url,
        json={"key": "value"},
    )
    out = api.fetch_json(url)
    assert isinstance(out, dict)
    assert out["key"] == "value"


@pytest.mark.remote
def test_fetch_dataset_json():
    start = 934000000
    end = 934100000
    out = api.fetch_dataset_json(start, end)
    assert not out["events"]
    assert {"S6", "history"}.issubset(set(out["runs"]))


@mock.patch("gwosc.api.v1.fetch_json")
def test_fetch_dataset_json_local(fetch):
    start = 934000000
    end = 934100000
    api.fetch_dataset_json(start, end)
    fetch.assert_called_with(losc_url(f"archive/{start}/{end}/json/"))


@pytest.mark.remote
def test_fetch_run_json():
    run = "S6"
    detector = "L1"
    start = 934000000
    end = 934100000
    out = api.fetch_run_json(run, detector, start, end)
    assert out["dataset"] == run
    assert out["GPSstart"] == start
    assert out["GPSend"] == end
    check_json_url_list(out["strain"])


@mock.patch("gwosc.api.v1.fetch_json")
def test_fetch_run_json_local(fetch):
    api.fetch_run_json("S6", "L1", 934000000, 934100000)
    fetch.assert_called_with(losc_url("archive/links/S6/L1/934000000/934100000/json/"))


@pytest.mark.remote
def test_fetch_cataloglist_json():
    out = api.fetch_cataloglist_json()
    assert "description" in out["GWTC-1-confident"]
    assert "url" in out["GWTC-1-confident"]


@mock.patch("gwosc.api.v1.fetch_json")
def test_fetch_cataloglist_json_local(fetch):
    api.fetch_cataloglist_json()
    fetch.assert_called_with(losc_url("eventapi/json/"))


@pytest.mark.remote
def test_fetch_catalog_json():
    out = api.fetch_catalog_json("GWTC-1-confident")
    events = out["events"]
    assert events["GW170817-v3"]["GPS"] == 1187008882.4


@mock.patch("gwosc.api.v1.fetch_json")
def test_fetch_catalog_json_local(fetch):
    api.fetch_catalog_json("GWTC-1-confident")
    fetch.assert_called_with(losc_url("eventapi/json/GWTC-1-confident/"))


@pytest.mark.remote
def test_fetch_filtered_events_json():
    out = api.fetch_filtered_events_json(select=["10 <= luminosity-distance <= 200"])
    events = out["events"]
    assert "GW190425-v1" in events


@mock.patch(
    "gwosc.api.v1.fetch_allowed_params_json",
    mock.MagicMock(return_value=_ALLOWED_PARAMS),
)
@mock.patch("gwosc.api.v1.fetch_json")
@pytest.mark.parametrize(
    "select",
    [
        "10 <= luminosity-distance <= 200",
        "10 =< luminosity-distance <= 200",
        "200 >= luminosity-distance >= 10",
        "10 <=  luminosity-distance   =< 200",
        "10<=luminosity-distance<=200",
        "  200 >=  luminosity-distance => 10   ",
        " 200 =>luminosity-distance=>10",
    ],
)
def test_fetch_filtered_events_json_local(fetch, select):
    api.fetch_filtered_events_json(select=[select])
    (called_url,), kwargs = fetch.call_args
    assert "max-luminosity-distance=200" in called_url
    assert "min-luminosity-distance=10" in called_url


@mock.patch(
    "gwosc.api.v1.fetch_allowed_params_json",
    mock.MagicMock(return_value=_ALLOWED_PARAMS),
)
@mock.patch("gwosc.api.v1.fetch_json")
@pytest.mark.parametrize(
    "select",
    [
        "100 <= luminosity-distance",
        "gps-time < 100",
        "gps-time  = > 100",
        "unknown-param <= 100",
        "c0mpl3telyR4nd-m",
    ],
)
def test_fetch_filtered_events_json_bad_local(fetch, select):
    with pytest.raises(ValueError, match="Could not parse"):
        api.fetch_filtered_events_json(select=[select])


@pytest.mark.remote
def test_fetch_event_json():
    out = api.fetch_event_json("GW150914")
    meta = out["events"]["GW150914-v3"]
    assert int(meta["GPS"]) == 1126259462
    assert meta["version"] == 3


@pytest.mark.remote
def test_fetch_event_json_with_hms_suffix():
    out = api.fetch_event_json("GW190930")
    assert list(out["events"].values())[0]["commonName"] == "GW190930_133541"


@pytest.mark.remote
def test_fetch_event_json_version():
    out = api.fetch_event_json("GW150914-v3")["events"]["GW150914-v3"]
    assert out["version"] == 3
    assert out["catalog.shortName"] == "GWTC-1-confident"


@pytest.mark.remote
def test_fetch_event_json_error_no_version():
    with pytest.raises(ValueError):
        api.fetch_event_json("GW150914-v3", version=1)


@pytest.mark.remote
def test_fetch_event_json_error_no_catalog():
    with pytest.raises(ValueError):
        api.fetch_event_json("GW150914-v3", catalog="test")


@pytest.mark.remote
def test_fetch_event_json_error_not_found():
    with pytest.raises(ValueError):
        api.fetch_event_json("blah")


@pytest.mark.remote
def test_fetch_event_json_exact_match():
    (data,) = list(api.fetch_event_json("GW190521")["events"].values())
    assert data["commonName"] == "GW190521"


@pytest.mark.remote
def test_fetch_event_json_error_multiple_names():
    with pytest.raises(ValueError) as exc:
        api.fetch_event_json("GW190828")
    assert str(exc.value).startswith("multiple events matched for 'GW190828'")



# Copyright (C) Cardiff University (2018-2021)
# SPDX-License-Identifier: MIT
import logging
from urllib.parse import urlencode

import requests

from gwosc import __version__

from .. import utils
from . import DEFAULT_URL

__all__ = [
    "fetch_json",
    "fetch_dataset_json",
    "fetch_run_json",
    "fetch_allowed_params_json",
    "fetch_cataloglist_json",
    "fetch_catalog_json",
    "fetch_allevents_json",
    "fetch_filtered_events_json",
    "fetch_event_json",
]

logger = logging.getLogger(__name__)

_MAX_GPS = 99999999999

#: Cache of downloaded blobs
JSON_CACHE = {}

# Timeout before a request is dropped
TIMEOUT = 10  # seconds

# -- JSON handling ------------------------------------------------------------


def fetch_json(url, **kwargs):
    """Fetch JSON data from a remote URL

    Parameters
    ----------
    url : `str`
        the remote URL to fetch

    **kwargs
        other keyword arguments are passed directly to :func:`requests.get`

    Returns
    ------
    data : `dict` or `list`
        the data fetched from ``url`` as parsed by
        :meth:`requests.Response.json`

    See also
    --------
    json.loads
        for details of the JSON parsing
    """
    try:
        return JSON_CACHE[url]
    except KeyError:
        logger.debug("fetching %s", url)
        client_headers = {"User-Agent": f"python-gwosc/{__version__}"}
        resp = requests.get(url, headers=client_headers, timeout=TIMEOUT, **kwargs)
        resp.raise_for_status()
        return JSON_CACHE.setdefault(
            url,
            resp.json(),
        )


# -- Run datasets -------------------------------------------------------------


def _dataset_url(start, end, host=DEFAULT_URL):
    return f"{host}/archive/{start:d}/{end:d}/json/"


def fetch_dataset_json(gpsstart, gpsend, host=DEFAULT_URL):
    """Returns the JSON metadata for all datasets matching the GPS interval

    Parameters
    ----------
    gpsstart : `int`
        the GPS start of the desired interval

    gpsend : `int`
        the GPS end of the desired interval

    host : `str`, optional
        the URL of the GWOSC host to query, defaults to
        https://gwosc.org

    Returns
    -------
    data : `dict` or `list`
        the JSON data retrieved from GWOSC and returned by `json.loads`
    """
    return fetch_json(_dataset_url(gpsstart, gpsend, host=host))


def _run_url(run, detector, start, end, host=DEFAULT_URL):
    return f"{host}/archive/links/{run}/{detector}/{start:d}/{end:d}/json/"


def fetch_run_json(run, detector, gpsstart=0, gpsend=_MAX_GPS, host=DEFAULT_URL):
    """Returns the JSON metadata for the given science run parameters

    Parameters
    ----------
    run : `str`
        the name of the science run, e.g. ``'O1'``

    detector : `str`
        the prefix of the GW detector, e.g. ``'L1'``

    gpsstart : `int`
        the GPS start of the desired interval

    gpsend : `int`
        the GPS end of the desired interval

    host : `str`, optional
        the URL of the GWOSC host to query, defaults to
        https://gwosc.org

    Returns
    -------
    data : `dict` or `list`
        the JSON data retrieved from GWOSC and returned by `json.loads`
    """
    return fetch_json(_run_url(run, detector, gpsstart, gpsend, host=host))


# -- EventAPI catalogs -------------------------------------------------------


def _allowed_params_url(host=DEFAULT_URL):
    return f"{host}/eventapi/json/query/params/"


def fetch_allowed_params_json(host=DEFAULT_URL):
    return fetch_json(_allowed_params_url(host=host))


def _eventapi_url(full=False, host=DEFAULT_URL):
    j = "jsonfull" if full else "json"
    return f"{host}/eventapi/{j}/"


def fetch_cataloglist_json(host=DEFAULT_URL):
    """Returns the JSON metadata for the catalogue list.

    Parameters
    ----------
    host : `str`, optional
        the URL of the GWOSC host to query

    Returns
    -------
    data : `dict` or `list`
        the JSON data retrieved from GWOSC and returned by
        :meth:`requests.Response.json`
    """
    return fetch_json(_eventapi_url(host=host))


def _catalog_url(catalog, host=DEFAULT_URL):
    return f"{_eventapi_url(host=host)}{catalog}/"


def fetch_catalog_json(catalog, host=DEFAULT_URL):
    """Returns the JSON metadata for the given catalogue

    Parameters
    ----------
    catalog : `str`
        the name of the event catalog, e.g. `GWTC-1-confident`

    host : `str`, optional
        the URL of the GWOSC host to query, defaults to
        https://gwosc.org

    Returns
    -------
    data : `dict` or `list`
        the JSON data retrieved from GWOSC and returnend by
        :meth:`requests.Response.json`
    """
    return fetch_json(_catalog_url(catalog, host=host))


# -- EventAPI events ---------------------------------------------------------


def _allevents_url(full=False, host=DEFAULT_URL):
    return f"{_eventapi_url(full=full, host=host)}allevents/"


def _has_jsonfull_allevents(host=DEFAULT_URL):
    return _allevents_url(full=True, host=host) in JSON_CACHE


def fetch_allevents_json(full=False, host=DEFAULT_URL):
    """Returns the JSON metadata for the allevents API

    Parameters
    ----------
    host : `str`, optional
        the URL of the GWOSC host to query, defaults to https://gwosc.org

    Returns
    -------
    data : `dict` or `list`
        the JSON data retrieved from GWOSC and returned by
        :meth:`requests.Response.json`
    """
    if full is None and _has_jsonfull_allevents(host=host):
        return fetch_json(_allevents_url(full=True, host=host))
    return fetch_json(_allevents_url(full=full, host=host))


def _fetch_allevents_event_json(
    event,
    catalog=None,
    version=None,
    full=False,
    host=DEFAULT_URL,
):
    """Returns the JSON metadata from the allevents view for a specific event

    The raw JSON data are packaged to look the same as if they came from
    a full event API query, i.e. nested under `'events`'.
    """
    allevents = fetch_allevents_json(full=full, host=host)["events"]
    matched = []

    def _match(keyvalue):
        dset, metadata = keyvalue
        if event not in {
            dset,
            metadata["commonName"],  # full name
            metadata["commonName"].split("_", 1)[0],  # GWYYMMDD prefix
        }:
            return
        if version is not None and metadata["version"] != version:
            return
        if catalog is not None and metadata["catalog.shortName"] != catalog:
            return
        return True

    # match datasets
    matched = list(filter(_match, allevents.items()))
    names = set(x[1]["commonName"] for x in matched)

    # one dataset has an exact name match, so discard everything else
    if event in names:
        matched = [x for x in matched if x[1]["commonName"] == event]
        names = set(x[1]["commonName"] for x in matched)

    # we have a winner!
    if len(names) == 1:
        key, meta = sorted(matched, key=lambda x: x[1]["version"])[-1]
        return {"events": {key: meta}}

    # raise error with the right message
    if len(names) > 1:
        raise ValueError(
            "multiple events matched for {!r}: '{}'".format(
                event,
                "', '".join(names),
            ),
        )
    msg = "failed to identify {} for event '{}'"
    if catalog is None:
        msg = msg.format("catalog", event)
        if version is not None:
            msg += f" at version {version}"
        raise ValueError(msg)
    msg = msg.format("version", event)
    if catalog is not None:
        msg += f" in catalog '{catalog}'"
    raise ValueError(msg)


def _query_events_url(select, host=DEFAULT_URL):
    query = urlencode(utils.select_to_query(select, host=host))
    return f"{host}/eventapi/json/query/show?{query}"


def fetch_filtered_events_json(select, host=DEFAULT_URL):
    """Return the JSON metadata for the events constrained by select

    Parameters
    ----------
    select : `list-like`
        a list of range constrains for the events.
        All ranges should have inclusive ends (<= and >= operators).

    host : `str`, optional
        the URL of the GWOSC host to query, defaults to
        https://gwosc.org

    Returns
    -------
    data : `dict` or `list`
        the JSON data retrieved from GWOSC and returnend by
        :meth:`requests.Response.json`

    Example
    -------
    >>> fetch_filtered_events_json(
    ...     select=[
    ...         "mass-1-source <= 5",
    ...         "mass-2-source =< 10",
    ...         "10 <= luminosity-distance <= 100",
    ...     ]
    ... )
    """
    return fetch_json(_query_events_url(select, host=host))


def _event_url(
    event,
    catalog=None,
    version=None,
    host=DEFAULT_URL,
):
    return list(
        _fetch_allevents_event_json(
            event,
            catalog=catalog,
            version=version,
            full=None,
            host=host,
        )["events"].values()
    )[0]["jsonurl"]


def fetch_event_json(
    event,
    catalog=None,
    version=None,
    host=DEFAULT_URL,
):
    """Returns the JSON metadata for the given event.

    By default, this function queries across all catalogs and all data-release
    versions, returning the highest available version, unless the
    ``version`` and/or ``catalog`` keywords are specified.

    Parameters
    ----------
    event : `str`
        the name of the event to query

    catalog : `str`, optional
        name of catalogue that hosts this event

    version : `int`, `None`, optional
        restrict query to a given data-release version

    host : `str`, optional
        the URL of the GWOSC host to query, defaults to
        https://gwosc.org

    Returns
    -------
    data : `dict` or `list`
        the JSON data retrieved from GWOSC and returned by `json.loads`
    """
    return fetch_json(
        _event_url(event, catalog=catalog, version=version, host=host),
    )



"""
:mod:`gwosc.api` provides the low-level interface functions
that handle direct requests to the GWOSC host.
"""

import os

#: The default GWOSC host URL
DEFAULT_URL = os.getenv("GWOSC_SERVER_URL", "https://gwosc.org")

from .v1 import *  # noqa

# Internal attributes for backwards compatibility
from .v1 import _MAX_GPS, _fetch_allevents_event_json  # noqa



# Copyright (C) Cardiff University (2018-2021)
# SPDX-License-Identifier: MIT

"""Tests for :mod:`gwosc.datasets`"""

import re
from unittest import mock

import pytest

from .. import datasets

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


@pytest.mark.remote
def test_find_datasets():
    sets = datasets.find_datasets()
    for dset in ("S6", "O1", "GW150914-v1", "GW170817-v3"):
        assert dset in sets
    assert "tenyear" not in sets
    assert "history" not in sets


@pytest.mark.remote
def test_find_datasets_detector():
    v1sets = datasets.find_datasets("V1")
    assert "GW170817-v3" in v1sets
    assert "GW150914-v1" not in v1sets

    assert datasets.find_datasets("X1", type="run") == []


@pytest.mark.remote
def test_find_datasets_type():
    runsets = datasets.find_datasets(type="run")
    assert "O1" in runsets
    run_regex = re.compile(
        r"\A([OS]\d+([a-z]|[A-Z]+)?|BKGW\d{6})(_\d+KHZ)?(_[RV]\d+)?\Z",
    )
    for dset in runsets:
        assert run_regex.match(dset)

    assert datasets.find_datasets(type="badtype") == []


@pytest.mark.remote
def test_find_datasets_segment():
    sets = datasets.find_datasets(segment=(1126051217, 1137254417))
    assert "GW150914-v1" in sets
    assert "GW170817" not in sets


@pytest.mark.remote
def test_find_datasets_match():
    assert "O1" not in datasets.find_datasets(match="GW")


@pytest.mark.remote
def test_find_datasets_event_version_detector():
    # this raises a ValueError with gwosc-0.5.0
    sets = datasets.find_datasets(type="event", version=1, detector="L1")
    assert "GW150914-v1" in sets
    assert "GW150914-v3" not in sets  # v3


@mock.patch("gwosc.datasets._run_datasets", return_value=[])
def test_find_datasets_warning(_):
    with pytest.warns(UserWarning):
        datasets.find_datasets(type="run", version=1)


@pytest.mark.remote
def test_event_gps():
    assert datasets.event_gps("GW170817") == 1187008882.4


@mock.patch(
    "gwosc.api.v2.fetch_event_version",
    return_value={
        "name": "GW150914",
        "gps": 12345.6,
        "something_else": 1,
    },
)
def test_event_gps_local(fetch):
    assert datasets.event_gps("GW150914") == 12345.6


@pytest.mark.remote
def test_event_segment():
    assert datasets.event_segment("GW170817") == (1187006835, 1187010931)


@mock.patch(
    "gwosc.api.v2.fetch_event_strain_data",
    mock.MagicMock(
        return_value=[
            {"gps_start": 0, "detector": "X1", "duration": 32},
            {"gps_start": 10, "detector": "Y1", "duration": 32},
        ]
    ),
)
def test_event_segment_local_without_det():
    assert datasets.event_segment("GW170817") == (0, 42)


@pytest.mark.remote
def test_event_at_gps():
    assert datasets.event_at_gps(1187008882) == "GW170817"
    with pytest.raises(ValueError) as exc:
        datasets.event_at_gps(1187008882, tol=0.1)
    assert str(exc.value) == "no event found within 0.1 seconds of 1187008882"


@mock.patch(
    "gwosc.api.v2.fetch_event_versions",
    mock.MagicMock(
        return_value=[
            {
                "name": "GW150914",
                "gps": 12345,
            }
        ]
    ),
)
def test_event_at_gps_local_found():
    assert datasets.event_at_gps(12345) == "GW150914"


@mock.patch(
    "gwosc.api.v2.fetch_event_versions",
    mock.MagicMock(return_value=[]),
)
def test_event_at_gps_local_not_found():
    with pytest.raises(ValueError):
        datasets.event_at_gps(12349)


@pytest.mark.remote
def test_event_detectors():
    assert datasets.event_detectors("GW150914") == {"H1", "L1"}
    assert datasets.event_detectors("GW170814") == {"H1", "L1", "V1"}


@mock.patch(
    "gwosc.api.v2.fetch_event_version",
    mock.MagicMock(
        return_value={
            "name": "GW150914",
            "gps": 12345.6,
            "detectors": ["A1", "B1"],
        }
    ),
)
def test_event_detectors_local():
    assert datasets.event_detectors("test") == {"A1", "B1"}


@pytest.mark.remote
def test_run_segment():
    assert datasets.run_segment("O1") == (1126051217, 1137254417)
    with pytest.raises(ValueError):
        datasets.run_segment("S7")


@mock.patch(
    "gwosc.api.v2.fetch_run",
    mock.MagicMock(return_value={"gps_start": 0, "gps_end": 1}),
)
def test_run_segment_local():
    assert datasets.run_segment("S1") == (0, 1)


@pytest.mark.remote
def test_run_at_gps():
    assert datasets.run_at_gps(1135136350) in {"O1", "O1_16KHZ"}
    with pytest.raises(ValueError) as exc:
        datasets.run_at_gps(0)
    assert str(exc.value) == "no run dataset found containing GPS 0"


@mock.patch(
    "gwosc.api.v2.fetch_runs",
    mock.MagicMock(
        return_value=[
            {"name": "S1", "gps_start": 0, "gps_end": 1},
            {"name": "S2", "gps_start": 3, "gps_end": 4},
        ]
    ),
)
def test_run_at_gps_local():
    assert datasets.run_at_gps(0) == "S1"
    with pytest.raises(ValueError):
        datasets.run_at_gps(10)


@pytest.mark.remote
def test_dataset_type():
    assert datasets.dataset_type("O1") == "run"
    assert datasets.dataset_type("GW150914-v1") == "event"
    assert datasets.dataset_type("GWTC-1-confident") == "catalog"
    with pytest.raises(ValueError):
        datasets.dataset_type("invalid")


@mock.patch(
    "gwosc.datasets.find_datasets",
    mock.MagicMock(side_effect=[["testrun"], [], ["testevent"], [], [], []]),
)
def test_dataset_type_local():
    assert datasets.dataset_type("testevent") == "event"
    with pytest.raises(ValueError):
        datasets.dataset_type("invalid")


@pytest.mark.remote
def test_query_events():
    events = datasets.query_events(select=["10 <= luminosity-distance <= 200"])
    assert "GW190425-v1" in events
    assert "GW190425-v2" in events
    assert "GW190425_081805-v3" in events


@mock.patch(
    "gwosc.api.v2.fetch_event_versions",
    mock.MagicMock(
        return_value=[
            {
                "name": "mock-event",
                "gps": 1240215503.0,
                "version": 1,
                "shortName": "mock-event-1",
            },
            {
                "name": "mock-event",
                "gps": 1240215503.0,
                "version": 2,
                "shortName": "mock-event-2",
            },
            {
                "name": "mock-event",
                "gps": 1240215503.0,
                "version": 3,
                "shortName": "mock-event-3",
            },
        ]
    ),
)
def test_query_events_local():
    events = datasets.query_events(
        select=["mass-1-source >= 1.4", "10 <= luminosity-distance <= 200"]
    )
    assert "mock-event-1" in events
    assert "mock-event-2" in events
    assert "mock-event-3" in events



# Copyright (C) Cardiff University (2018-2021)
# SPDX-License-Identifier: MIT

"""Helpers for tests"""

import os

import pytest

from ..api import (
    fetch_event_json,
)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

# unset these variables to allow the tests to run with pybuild
os.environ.pop("http_proxy", None)
os.environ.pop("https_proxy", None)


def _event_strain(name, **kwargs):
    return list(fetch_event_json(name, **kwargs)["events"].values())[0]["strain"]


@pytest.fixture(scope="module")
def gw150914_strain():
    return _event_strain("GW150914", version=3)


@pytest.fixture
def mock_strain():
    return [
        {
            "url": "X-X1_LOSC_TEST_4_V1-0-32.ext",
            "GPSstart": 0,
            "duration": 32,
            "detector": "X1",
            "sampling_rate": 4096,
        },
        {
            "url": "X-X1_LOSC_TEST_4_V1-32-32.ext",
            "GPSstart": 32,
            "duration": 32,
            "detector": "X1",
            "sampling_rate": 4096,
        },
        {
            "url": "Y-Y1_LOSC_TEST2_4_V1-0-32.ext",
            "GPSstart": 0,
            "duration": 32,
            "detector": "Y1",
            "sampling_rate": 4096,
        },
        {
            "url": "Y-Y1_LOSC_TEST2_16_V2-0-32.ext",
            "GPSstart": 0,
            "duration": 32,
            "detector": "Y1",
            "sampling_rate": 16384,
        },
    ]


@pytest.fixture
def mock_urls(mock_strain):
    return [u["url"] for u in mock_strain]



# Copyright (C) Cardiff University (2018-2021)
# SPDX-License-Identifier: MIT

import re
from os.path import basename

from .api import DEFAULT_URL
from .api.v2 import fetch_allowed_params

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

_ALLOWED_OPS = set((">=", "=>", "<=", "=<"))


def url_segment(url):
    """Return the GPS segment covered by a URL following T050017

    Parameters
    ---------
    filename : `str`
        the path name of a file

    Returns
    -------
    segment : `tuple` of `int`
        the ``[start, stop)`` GPS interval covered by the given URL

    Notes
    -----
    `LIGO-T050017 <https://dcc.ligo.org/LIGO-T050017/public>`_ declares
    a filenaming convention that includes documenting the GPS start integer
    and integer duration of a file, see that document for more details.
    """
    base = basename(url)
    _, _, s, e = base.split("-")
    s = int(s)
    e = int(e.split(".")[0])
    return s, s + e


def url_overlaps_segment(url, segment):
    """Returns True if a URL overlaps a [start, stop) GPS interval

    Parameters
    ----------
    url : `str`
        the URL of a file

    segment : `tuple` of `int`
        the ``[start, stop)`` GPS interval to check against

    Returns
    -------
    overlap? : `bool`
        `True` if the GPS interval for the URL matches that given,
        otherwise `False`
    """
    useg = url_segment(url)
    return segments_overlap(useg, segment)


def urllist_extent(urls):
    """Returns the GPS `[start, end)` interval covered by a list or URLs

    Parameters
    ----------
    urls : `iterable` of `str`
        the list of URLs

    Returns
    -------
    a, b : 2-`tuple` of int`
        the GPS extent of the URL list
    """
    segs = map(url_segment, urls)
    starts, ends = zip(*segs)
    return min(starts), max(ends)


def strain_extent(strain):
    """Returns the GPS `[start, end)` interval covered by a strain meta dict"""
    starts, ends = zip(
        *[(meta["GPSstart"], meta["GPSstart"] + meta["duration"]) for meta in strain]
    )
    return min(starts), max(ends)


def full_coverage(urls, segment):
    """Returns True if the list of URLS completely covers a GPS interval

    The URL list is presumed to be contiguous, so this just checks that
    the first URL (by GPS timestamp) and the last URL can form a segment
    that overlaps that given.
    """
    if not urls:
        return False
    # sort URLs by GPS timestamp
    a, b = urllist_extent(urls)
    return a <= segment[0] and b >= segment[1]


def segments_overlap(a, b):
    """Returns True if GPS segment ``a`` overlaps GPS segment ``b``"""
    return (a[1] > b[0]) and (a[0] < b[1])


def parse_two_ops(compiled_m, host=DEFAULT_URL):
    allowed_params = fetch_allowed_params(host=host)
    md = compiled_m.groupdict()
    op1, op2 = md["op1"], md["op2"]
    if not set((op1, op2)).issubset(_ALLOWED_OPS):
        raise ValueError("Could not parse select string.\n" "Unknown operators.")
    param, val1, val2 = md["param"], md["val1"], md["val2"]
    if param not in allowed_params:
        raise ValueError(
            "Could not parse select string.\n"
            f"Unrecognized parameter: {param}.\n"
            f"Use one of:\n{allowed_params}"
        )
    queries = []
    if ">" in op1:
        queries.append((f"max-{param}", val1))
    if "<" in op1:
        queries.append((f"min-{param}", val1))
    if ">" in op2:
        queries.append((f"min-{param}", val2))
    if "<" in op2:
        queries.append((f"max-{param}", val2))
    return queries


def parse_one_op(compiled_m, host=DEFAULT_URL):
    allowed_params = fetch_allowed_params(host=host)
    md = compiled_m.groupdict()
    op = md["op"]
    if not set((op,)).issubset(_ALLOWED_OPS):
        raise ValueError("Could not parse select string.\n" "Unknown operator.")
    param, val = md["param"], md["val"]
    if param not in allowed_params:
        raise ValueError(
            f"Could not parse select string.\n"
            f"Unrecognized parameter: {param}.\n"
            f"Use one of:\n{allowed_params}"
        )
    queries = []
    if ">" in op:
        queries.append((f"min-{param}", val))
    if "<" in op:
        queries.append((f"max-{param}", val))
    return queries


def select_to_query(select, host=DEFAULT_URL):
    """Parse select string and translate into URL GET parameters"""

    # Captures strings of the form `1.44 <= param <= 5.0`
    two_ops = re.compile(
        r"^\s*(?P<val1>[\d.+-eE]+)\s*(?P<op1>[<>=]{2})\s*(?P<param>[\w-]+)"
        r"\s*(?P<op2>[<>=]+)\s*(?P<val2>[\d.+-eE]+)\s*$"
    )
    # Captures strings of the form `param <= 5.0`
    one_op = re.compile(
        r"^\s*(?P<param>[\w-]+)\s*(?P<op>[<>=]{2})\s*(?P<val>[\d.+-eE]+)\s*$"
    )

    queries = []
    for s in select:
        for regex, parse in (
            (one_op, parse_one_op),
            (two_ops, parse_two_ops),
        ):
            m = regex.match(s)
            if m is not None:
                queries.extend(parse(m, host=host))
                break
        else:
            raise ValueError(f"Could not parse select string: {s}")
    return queries



# Copyright (C) Cardiff University (2018-2021)
# SPDX-License-Identifier: MIT

"""
`gwosc.timeline` provides functions to find segments for a given dataset.

You can search for Timeline segments, based on a flag name, and a
GPS time interval as follows:

>>> from gwosc.timeline import get_segments
>>> get_segments("H1_DATA", 1126051217, 1126151217)
[(1126073529, 1126114861), (1126121462, 1126123267), (1126123553, 1126126832), (1126139205, 1126139266), (1126149058, 1126151217)]

The output is a `list` of ``[start, end)`` 2-tuples which each
represent a semi-open time interval.

For documentation on what flags are available, for example for the O1
science run, see `the O1 data release
page <https://gwosc.org/O1/>`__ (*Data Quality*).
"""  # noqa: E501

import warnings
from operator import itemgetter

from . import datasets
from .api import DEFAULT_URL
from .api import v2 as apiv2


def get_segments(flag, start, end, host=DEFAULT_URL):
    """Return the [start, end) GPS segments for this flag

    Parameters
    ----------
    flag : `str`
        name of flag, e.g. ``'H1_DATA'``

    start : `int`
        the GPS start time of your query

    end : `int`
        the GPS end time of your query

    host : `str`, optional
        the URL of the remote GWOSC server

    Returns
    -------
    segments : `list` of `(int, int)` tuples
        a list of `[a, b)` GPS segments
    """
    return list(
        map(
            lambda s: tuple(s.values()),
            apiv2.fetch_segments(flag, start, end, host=host),
        )
    )


def timeline_url(flag, start, end, host=DEFAULT_URL):
    """Returns the Timeline JSON URL for a flag name and GPS interval"""
    warnings.warn(
        "timeline_url function will be removed in future versions.",
        DeprecationWarning,
        stacklevel=2,
    )
    detector = flag.split("_", 1)[0]
    dataset = _find_dataset(start, end, detector, host=host)
    return f"{host}/timeline/segments/json/" f"{dataset}/{flag}/{start}/{end - start}/"


def _find_dataset(start, end, detector, host=DEFAULT_URL):
    warnings.warn(
        "_find_dataset function will be removed in future versions.",
        DeprecationWarning,
        stacklevel=2,
    )
    duration = end - start
    epochs = []

    for run in datasets._iter_datasets(
        type="run",
        detector=detector,
        segment=(start, end),
        host=host,
    ):
        segment = datasets.run_segment(run, host=host)
        overlap = min(end, segment[1]) - max(start, segment[0])
        epochs.append((run, duration - overlap))

    if not epochs:
        raise ValueError(f"No datasets found matching [{start}, {end})")

    return sorted(epochs, key=itemgetter(1, 0))[0][0]



# Copyright (C) Cardiff University (2018-2021)
# SPDX-License-Identifier: MIT

"""Tests for :mod:`gwosc.utils`"""

import pytest

from .. import utils

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def test_url_segment():
    seg = utils.url_segment("X-TEST-123-456.ext")
    assert seg == (123, 579)


@pytest.mark.parametrize(
    "url, segment, result",
    [
        ("A-B-10-1.ext", (0, 10), False),
        ("A-B-10-1.ext", (5, 11), True),
        ("A-B-10-1.ext", (10, 15), True),
        ("A-B-10-1.ext", (11, 15), False),
    ],
)
def test_url_overlaps_segment(url, segment, result):
    assert utils.url_overlaps_segment(url, segment) is result


@pytest.mark.remote
@pytest.mark.parametrize(
    "segment, result",
    [
        ((0, 64), True),
        ((1, 63), True),
        ((-1, 63), False),
        ((-1, 64), False),
        ((0, 65), False),
        ((1, 65), False),
        ((-1, 64), False),
    ],
)
def test_full_coverage(mock_urls, segment, result):
    assert utils.full_coverage(mock_urls, segment) is result


def test_full_coverage_empty():
    assert utils.full_coverage([], (0, 1)) is False


@pytest.mark.parametrize(
    "seg1, seg2, result",
    [
        ((10, 11), (0, 10), False),
        ((10, 11), (5, 11), True),
        ((10, 11), (10, 15), True),
        ((10, 11), (11, 15), False),
    ],
)
def test_segments_overlap(seg1, seg2, result):
    assert utils.segments_overlap(seg1, seg2) is result



# Copyright (C) Cardiff University (2018-2021)
# SPDX-License-Identifier: MIT
import logging
import time
import warnings
from urllib import parse

import requests

from gwosc import __version__

from . import DEFAULT_URL

logger = logging.getLogger(__name__)

#: Cache of downloaded blobs
JSON_CACHE = {}

#: Number of pages to fetch that trigger a warning
LARGE_NUM_PAGES_WARN = 100
# Number of request retries after a 429 code, before we raise an exception
MAX_RETRIES = 10
# Default wait time before we retry, if the `Retry-After` header key in the
# response is not set, or if it can't be parsed as an integer
RETRY_DEFAULT_WAIT_TIME = 1  # seconds
# Timeout before a request is dropped
TIMEOUT = 10  # seconds


# -- JSON handling ------------------------------------------------------------


def fetch_json(url, **kwargs):
    """Fetch JSON data from a remote URL

    Parameters
    ----------
    url : `str`
        the remote URL to fetch

    **kwargs
        other keyword arguments are passed directly to :func:`requests.get`

    Returns
    ------
    data : `dict` or `list`
        the data fetched from ``url`` as parsed by
        :meth:`requests.Response.json`

    See also
    --------
    json.loads
        for details of the JSON parsing
    """
    if url in JSON_CACHE.keys():
        return JSON_CACHE[url]

    logger.debug("fetching %s", url)
    client_headers = {"User-Agent": f"python-gwosc/{__version__}"}
    resp = requests.get(url, headers=client_headers, timeout=TIMEOUT, **kwargs)
    retry_attempt = 0
    while resp.status_code == 429 and retry_attempt < MAX_RETRIES:
        # we made too many requests per second
        try:
            retry_after = int(resp.headers.get("Retry-After", RETRY_DEFAULT_WAIT_TIME))
        except ValueError:
            retry_after = RETRY_DEFAULT_WAIT_TIME
        time.sleep(retry_after)
        resp = requests.get(url, headers=client_headers, timeout=TIMEOUT, **kwargs)
        retry_attempt = retry_attempt + 1
    resp.raise_for_status()
    return JSON_CACHE.setdefault(url, resp.json())


def produce_fetched_objects(url):
    current_page = fetch_json(url)
    num_pages = current_page["num_pages"]
    if num_pages > LARGE_NUM_PAGES_WARN:
        warn_message = (
            f"Your request will need to fetch for {num_pages} pages. "
            "Try to constraint your request if possible."
        )
        warnings.warn(warn_message)
        logger.warning(warn_message)
    yield from current_page["results"]
    # Iterate over the rest of the pages
    next_page = current_page["next"]
    while next_page is not None:
        current_page = fetch_json(next_page)
        next_page = current_page["next"]
        yield from current_page["results"]


def fetch_run_strain_files(
    run=None,
    detector=None,
    start=None,
    end=None,
    sample_rate=None,
    host=DEFAULT_URL,
):
    """Return strain file objects from bulk-data releases.

    Parameters
    ----------
    run : `str`, optional
        the ID of a run, e.g. ``'O1'``

    detector : `str`, optional
        the prefix of the GW detector, e.g. ``'L1'``

    start : `int`, optional
        the GPS start of the desired interval

    end : `int`, optional
        the GPS end of the desired interval

    sample_rate : `int`, optional
        the sample rate of the strain file data, either 4096 or 16384 [Hz]

    host : `str`, optional
        the URL of the GWOSC host to query, defaults to
        https://gwosc.org

    Returns
    -------
    data : `iterable[dict]`
        An iterable of strain file dictionaries.
    """

    if run is not None:
        if "KHZ_R1" in run:
            run = run.split("_")[0]
        strain_url = f"{host}/api/v2/runs/{run}/strain-files"
    else:
        strain_url = f"{host}/api/v2/strain-files"

    # Prepare query parameters
    start = None if start is None else int(start)
    end = None if end is None else int(end)
    sample_rate = {4096: 4, 16_384: 16}.get(sample_rate, sample_rate)
    query_params = {
        k: v
        for k, v in {
            "detector": detector,
            "start": start,
            "stop": end,
            "sample-rate": sample_rate,
        }.items()
        if v is not None
    }
    query_string = parse.urlencode(query_params)

    if query_string:
        strain_url = f"{strain_url}?{query_string}"

    yield from produce_fetched_objects(strain_url)


def fetch_event_strain_data(
    event,
    detector=None,
    version=None,
    catalog=None,
    sample_rate=None,
    duration=None,
    format="hdf5",
    host=DEFAULT_URL,
):
    """Return strain file objects from single-event releases.

    Parameters
    ----------
    event : `str` or `None`
        the ID of an event, e.g. ``'GW150914'``

    detector : `str`
        the prefix of the GW detector, e.g. ``'L1'``

    version : `int`
        the version number of the requested event

    catalog : `str`
        the catalog in which the requested event appears

    sample_rate : `int`
        the sample rate of the strain file data, either 4096 or 16384 [Hz]

    duration : `int`
        the duration of the strain file, 32 or 4096 [s]

    format : `str`
        the file format of the strain file.
        One of ``'hdf'``, ``'gwf'``, ``'txt'``

    host : `str`, optional
        the URL of the GWOSC host to query, defaults to
        https://gwosc.org

    Returns
    -------
    data : `iterable[dict]`
        An iterable of strain file dictionaries.
    """

    # Prepare event ID
    event_id = event
    if "-v" not in event:
        if version is not None:
            event_id = f"{event}-v{version}"
        elif catalog is not None:
            event_id = f"{event}@{catalog}"

    strain_url = f"{host}/api/v2/event-versions/{event_id}/strain-files"

    # Prepare query parameters
    format = "hdf" if format == "hdf5" else format
    sample_rate = {4096: 4, 16_384: 16}.get(sample_rate, sample_rate)
    query_params = {
        k: v
        for k, v in {
            "detector": detector,
            "sample-rate": sample_rate,
            "file-format": format,
            "duration": duration,
        }.items()
        if v is not None
    }
    query_string = parse.urlencode(query_params)

    if query_string:
        strain_url = f"{strain_url}?{query_string}"

    yield from produce_fetched_objects(strain_url)


def fetch_segments(flag, start, end, host=DEFAULT_URL):
    """Return segment dictionaries in the (start, end) GPS time interval.

    Parameters
    ----------
    flag : `str`
        name of flag, e.g. ``'H1_DATA'``

    start : `int`
        the GPS start time of your query

    end : `int`
        the GPS end time of your query

    host : `str`, optional
        the URL of the GWOSC host to query, defaults to
        https://gwosc.org

    Returns
    -------
    data : `iterable[dict]`
        An iterable of segment dictionaries.
    """
    segments_url = f"{host}/api/v2/timelines/{flag}/segments?start={start}&stop={end}"
    yield from produce_fetched_objects(segments_url)


def fetch_runs(host=DEFAULT_URL):
    """Return a list of all past runs

    Parameters
    ----------
    host : `str`, optional
        the URL of the GWOSC host to query, defaults to
        https://gwosc.org

    Returns
    -------
    data : `iterable[dict]`
        An iterable of runs.
    """
    yield from produce_fetched_objects(f"{host}/api/v2/runs")


def fetch_catalogs(host=DEFAULT_URL):
    """Returns a list with all catalogs

    Parameters
    ----------
    host : `str`, optional
        the URL of the GWOSC host to query, defaults to
        https://gwosc.org

    Returns
    -------
    data : `iterable[dict]`
        A list of catalogs
    """
    yield from produce_fetched_objects(f"{host}/api/v2/catalogs")


def fetch_event_versions(
    name=None, segment=None, catalogs=None, select=None, host=DEFAULT_URL
):
    """Returns an event

    Parameters
    ----------
    name: `str`, optional
        a full or partial name for an event

    segment: `tuple`, optional
        a gps time tuple (start, end) to restrict the search

    catalogs: `str`, `iterable`, optional
        a single catalog name or a list of catalog names

    select: `dict`, optional
        a dictionary with query parameters, e.g. {'min-p-astro': 0.5}

    host: `str`, optional
        the URL of the GWOSC host to query, defaults to
        https://gwosc.org

    Returns
    -------
    data : `iterable[dict]`
        A list of event version dictionaries
    """
    events_url = f"{host}/api/v2/event-versions"

    query = {}
    if name is not None:
        query["name-contains"] = name

    if segment is not None:
        query["min-gps-time"], query["max-gps-time"] = segment

    if catalogs is not None:
        if isinstance(catalogs, str):
            query["release"] = catalogs
        else:
            if len(catalogs) > 0:
                query["release"] = ",".join(catalogs)

    if select is not None:
        query = {**query, **select}

    query_string = parse.urlencode(query)

    if query_string:
        events_url = f"{events_url}?{query_string}"

    yield from produce_fetched_objects(events_url)


def fetch_run(run, host=DEFAULT_URL):
    """Return a run detail

    Parameters
    ----------

    run : `str`
        the name of the run, e.g. O1

    host : `str`, optional
        the URL of the GWOSC host to query, defaults to
        https://gwosc.org

    Returns
    -------
    data : `dict`
        A dictionary with the run detail
    """
    try:
        run = fetch_json(f"{host}/api/v2/runs/{run}")
    except requests.HTTPError:
        raise ValueError(f"Run '{run}' not found.")
    return run


def fetch_event_version(event, catalog=None, version=None, host=DEFAULT_URL):
    """Returns an event

    Parameters
    ----------
    event : `str`
        the name of the event

    catalog : `str`, optional
        name of catalogue that hosts this event

    version : `int`, `None`, optional
        the version of the data release to use,
        defaults to the highest available version

    host : `str`, optional
        the URL of the GWOSC host to query, defaults to
        https://gwosc.org

    Returns
    -------
    data : `dict`
        A dictionary with the event detail
    """
    event_string = event
    if "-v" in event:
        event_string = event
    elif version is not None:
        event_string = f"{event}-v{version}"
    elif catalog is not None:
        event_string = f"{event}@{catalog}"
    event_url = f"{host}/api/v2/event-versions/{event_string}"

    return fetch_json(event_url)


def fetch_allowed_params(host=DEFAULT_URL):
    return list(produce_fetched_objects(f"{host}/api/v2/default-parameters"))



# Copyright (C) Cardiff University (2018-2021)
# SPDX-License-Identifier: MIT

"""Tests for :mod:`gwosc.api.v2`"""

from unittest.mock import patch

import pytest

from ..api import v2 as apiv2

__author__ = "Martin Beroiz <martin.beroiz@ligo.org>"


def test_cache():
    apiv2.JSON_CACHE["something"] = "cached_value"
    assert apiv2.fetch_json("something") == "cached_value"


@patch("gwosc.api.v2.fetch_json")
def test_produce_fetched_objects_pagination(mock_fetch):
    mock_fetch.side_effect = [
        {
            "next": "http://dummy_url.com?page=2",
            "results": [1, 2, 3],
            "num_pages": 10,
        },
        {
            "next": None,
            "results": [4, 5, 6],
            "num_pages": 10,
        },
    ]
    result = list(apiv2.produce_fetched_objects("mock_url"))
    assert result == [1, 2, 3, 4, 5, 6]


@patch("gwosc.api.v2.fetch_json")
def test_produce_fetched_objects_toomanypages(mock_fetch):
    mock_fetch.side_effect = [
        {
            "next": "http://dummy_url.com?page=2",
            "results": [1, 2, 3],
            "num_pages": 1000,
        },
        {
            "next": None,
            "results": [4, 5, 6],
            "num_pages": 1000,
        },
    ]
    with pytest.warns(UserWarning, match="1000 pages"):
        result = list(apiv2.produce_fetched_objects("mock_url"))
        assert result == [1, 2, 3, 4, 5, 6]


@pytest.mark.remote
def test_fetch_run_strain_files():
    s5_files = list(
        apiv2.fetch_run_strain_files(
            detector="H1", run="S5", start=826175488, end=826200064
        )
    )
    assert len(s5_files) == 4
    o1_files = list(
        apiv2.fetch_run_strain_files(
            detector="L1", run="O1", start=1127407616, end=1127420000
        )
    )
    assert len(o1_files) == 4


@pytest.mark.remote
def test_fetch_event_strain_data():
    gw_files = list(
        apiv2.fetch_event_strain_data(
            "GW150914", version=3, sample_rate=4096, format="txt"
        )
    )
    assert len(gw_files) == 4
    for afile in gw_files:
        assert afile["sample_rate_kHz"] == 4
        assert afile["file_format"].lower() == "txt"


@pytest.mark.remote
def test_fetch_segments():
    segs = list(apiv2.fetch_segments("H1_DATA", start=932540000, end=932560000))
    assert len(segs) == 2


@pytest.mark.remote
def test_fetch_runs():
    assert "S6" in set(run["name"] for run in apiv2.fetch_runs())


@pytest.mark.remote
def test_fetch_catalogs():
    catalogs = set(cat["name"] for cat in apiv2.fetch_catalogs())
    assert "GWTC" in catalogs
    assert "GWTC-1-confident" in catalogs


@pytest.mark.remote
def test_fetch_event_versions():
    event = list(apiv2.fetch_event_versions(name="GW150914"))[0]
    assert event["name"] == "GW150914"

    events = list(apiv2.fetch_event_versions(catalogs="GWTC-1-confident"))
    assert len(events) == 11

    event = list(apiv2.fetch_event_versions(segment=(1126259450, 1126259470)))[0]
    assert event["name"] == "GW150914"

    events = list(
        apiv2.fetch_event_versions(
            select={
                "max-mass-1-source": 5,
                "min-p-astro": 0.5,
            }
        )
    )[0]
    assert len(events) > 0


def test_fetch_run():
    run = apiv2.fetch_run("S5")
    assert run["name"] == "S5"


def test_fetch_event_version():
    event = apiv2.fetch_event_version("GW150914", version=3)
    assert "150914" in event["name"]
    assert event["version"] == 3

    event = apiv2.fetch_event_version("GW150914", catalog="GWTC")
    assert "150914" in event["name"]
    assert "GWTC" in event["catalog"]
    assert event["catalog"] == "GWTC-1-confident"


def test_fetch_allowed_params():
    params = apiv2.fetch_allowed_params()
    assert len(params) > 0



# Copyright (C) Cardiff University (2018-2021)
# SPDX-License-Identifier: MIT

"""GWOSC: a python interface to the GW Open Science data archive"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__license = "MIT"

try:
    from ._version import version as __version__
except ModuleNotFoundError:  # development mode
    __version__ = ""



# Copyright (C) Cardiff University (2018-2021)
# SPDX-License-Identifier: MIT

"""Tests for :mod:`gwosc.locate`"""

import os.path
from pathlib import Path

import pytest

from .. import (
    locate,
    utils,
)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


@pytest.mark.remote
def test_get_urls():
    # test simple fetch for S6 data returns only files within segment
    detector = "L1"
    start = 934000000
    end = 934100000
    span = (start, end)
    urls = locate.get_urls(detector, start, end)
    for url in urls:
        assert os.path.basename(url).startswith(f"{detector[0]}-{detector}")
        assert utils.segments_overlap(utils.url_segment(url), span)

    # test fetch for GW170817 data
    assert (
        len(
            locate.get_urls(
                "L1",
                1187007040,
                1187009088,
                dataset="GW170817-v3",
            )
        )
        == 2
    )

    # test for O1 data
    assert len(locate.get_urls("L1", 1135136228, 1135140324)) == 2

    # assert no hits raises exception
    with pytest.raises(ValueError):  # no data in 1980
        locate.get_urls(detector, 0, 1)
    with pytest.raises(ValueError):  # no Virgo data for S6
        locate.get_urls("V1", start, end)


@pytest.mark.remote
def test_get_urls_version():
    """Regression test against version not being respected in get_urls"""
    urls = locate.get_urls("L1", 1187008877, 1187008887, version=2)
    assert Path(urls[0]).name == "L-L1_LOSC_CLN_4_V1-1187007040-2048.hdf5"


@pytest.mark.remote
def test_get_event_urls():
    urls = locate.get_event_urls("GW150914-v3", sample_rate=4096)
    assert len(urls) == 4
    for url in urls:
        assert "_4KHZ" in url


@pytest.mark.remote
def test_get_event_urls_segment():
    urls = locate.get_event_urls(
        "GW150914-v1",
        start=1126257415,
        end=1126257425,
    )
    assert len(urls) == 2
    for url in urls:  # check that these are the 4096-second files
        assert "-4096." in url


@pytest.mark.remote
def test_get_urls_gw170104():
    # check that we can find the right URL from an event dataset for
    # a GPS time that doesn't overlap with the event, and ends before
    # the start of the 32-second files (this used to not work)
    urls = locate.get_urls("L1", 1167558912.6, 1167559012.6)
    assert list(map(os.path.basename, urls)) == [
        "L-L1_GWOSC_4KHZ_R1-1167557889-4096.hdf5",
    ]
